###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import numpy as np
import unittest
from vistrails.tests.utils import execute, intercept_results

from vistrails.packages.sklearn.init import (Digits, Iris, TrainTestSplit,
                                             Predict, Score, Transform,
                                             CrossValScore, _modules,
                                             GridSearchCV)
from vistrails.packages.sklearn import identifier

from sklearn.metrics import f1_score


def class_by_name(name):
    """Returns an autogenerated class from _modules from a string name."""
    for module in _modules:
        if module.__name__ == name:
            return module


class TestSklearn(unittest.TestCase):
    def test_digits(self):
        # check that the digits dataset can be loaded
        with intercept_results(Digits, 'data', Digits, 'target') as (data, target):
            self.assertFalse(execute([
                ('datasets|Digits', identifier, [])
            ]))
        data = np.vstack(data)
        target = np.hstack(target)
        self.assertEqual(data.shape, (1797, 64))
        self.assertEqual(target.shape, (1797,))

    def test_iris(self):
        # check that the iris dataset can be loaded
        with intercept_results(Iris, 'data', Iris, 'target') as (data, target):
            self.assertFalse(execute([
                ('datasets|Iris', identifier, [])
            ]))
        data = np.vstack(data)
        target = np.hstack(target)
        self.assertEqual(data.shape, (150, 4))
        self.assertEqual(target.shape, (150,))

    def test_train_test_split(self):
        # check that we can split the iris dataset
        with intercept_results(TrainTestSplit, 'training_data', TrainTestSplit,
                               'training_target', TrainTestSplit, 'test_data',
                               TrainTestSplit, 'test_target') as results:
            X_train, y_train, X_test, y_test = results
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('cross-validation|TrainTestSplit', identifier,
                     [('test_size', [('Integer', '50')])])
                ],
                [
                    (0, 'data', 1, 'data'),
                    (0, 'target', 1, 'target')
                ]
            ))
        X_train = np.vstack(X_train)
        X_test = np.vstack(X_test)
        y_train = np.hstack(y_train)
        y_test = np.hstack(y_test)
        self.assertEqual(X_train.shape, (100, 4))
        self.assertEqual(X_test.shape, (50, 4))
        self.assertEqual(y_train.shape, (100,))
        self.assertEqual(y_test.shape, (50,))

    def test_classifier_training_predict(self):
        with intercept_results(Predict, 'prediction', Predict,
                               'decision_function', TrainTestSplit, 'test_target',
                               Score, 'score') as results:
            y_pred, decision_function, y_test, score = results
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('cross-validation|TrainTestSplit', identifier,
                     [('test_size', [('Integer', '50')])]),
                    ('classifiers|LinearSVC', identifier, []),
                    ('Predict', identifier, []),
                    ('Score', identifier, []),
                    # use custom metric
                    ('Score', identifier,
                     [('metric', [('String', 'f1')])]),

                ],
                [
                    # train test split
                    (0, 'data', 1, 'data'),
                    (0, 'target', 1, 'target'),
                    # fit LinearSVC on training data
                    (1, 'training_data', 2, 'training_data'),
                    (1, 'training_target', 2, 'training_target'),
                    # predict on test data
                    (2, 'model', 3, 'model'),
                    (1, 'test_data', 3, 'data'),
                    # score test data
                    (2, 'model', 4, 'model'),
                    (1, 'test_data', 4, 'data'),
                    (1, 'test_target', 4, 'target'),
                    # f1 scorer
                    (2, 'model', 5, 'model'),
                    (1, 'test_data', 5, 'data'),
                    (1, 'test_target', 5, 'target')
                ]
            ))
        y_pred = np.hstack(y_pred)
        decision_function = np.vstack(decision_function)
        y_test = np.hstack(y_test)
        # unpack the results from the two scorers
        score_acc, score_f1 = score
        self.assertEqual(y_pred.shape, (50,))
        self.assertTrue(np.all(np.unique(y_pred) == np.array([0, 1, 2])))
        self.assertEqual(decision_function.shape, (50, 3))
        # some accuracy
        self.assertTrue(np.mean(y_test == y_pred) > .8)
        # score is actually the accuracy
        self.assertEqual(np.mean(y_test == y_pred), score_acc)
        # f1 score is actually f1 score
        self.assertEqual(f1_score(y_test, y_pred), score_f1)

    def test_transformer_supervised_transform(self):
        # test feature selection
        with intercept_results(Transform, 'transformed_data') as (transformed_data,):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('feature_selection|SelectKBest', identifier,
                        [('k', [('Integer', '2')])]),
                    ('Transform', identifier, [])
                ],
                [
                    (0, 'data', 1, 'training_data'),
                    (0, 'target', 1, 'training_target'),
                    (1, 'model', 2, 'model'),
                    (0, 'data', 2, 'data')
                ]
            ))
        transformed_data = np.vstack(transformed_data)
        self.assertEqual(transformed_data.shape, (150, 2))

    def test_transformer_unsupervised_transform(self):
        # test PCA
        with intercept_results(Transform, 'transformed_data') as (transformed_data,):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('decomposition|PCA', identifier,
                        [('n_components', [('Integer', '2')])]),
                    ('Transform', identifier, [])
                ],
                [
                    (0, 'data', 1, 'training_data'),
                    (1, 'model', 2, 'model'),
                    (0, 'data', 2, 'data')
                ]
            ))
        transformed_data = np.vstack(transformed_data)
        self.assertEqual(transformed_data.shape, (150, 2))

    def test_manifold_learning(self):
        # test Isomap
        with intercept_results(class_by_name("Isomap"), 'transformed_data') as (transformed_data,):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('manifold|Isomap', identifier, []),
                ],
                [
                    (0, 'data', 1, 'training_data'),
                ]
            ))
        transformed_data = np.vstack(transformed_data)
        self.assertEqual(transformed_data.shape, (150, 2))

    def test_cross_val_score(self):
        # chech that cross_val score of LinearSVC has the right length
        with intercept_results(CrossValScore, 'scores') as (scores,):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('classifiers|LinearSVC', identifier, []),
                    ('cross-validation|CrossValScore', identifier, []),
                ],
                [
                    (0, 'data', 2, 'data'),
                    (0, 'target', 2, 'target'),
                    (1, 'model', 2, 'model')
                ]
            ))
        scores = np.hstack(scores)
        self.assertEqual(scores.shape, (3,))
        self.assertTrue(np.mean(scores) > .8)

    def test_gridsearchcv(self):
        # check that gridsearch on DecisionTreeClassifier does the right number of runs
        # and gives the correct result.
        with intercept_results(GridSearchCV, 'scores', GridSearchCV,
                               'best_parameters') as (scores, parameters):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('classifiers|DecisionTreeClassifier', identifier, []),
                    ('GridSearchCV', identifier,
                     [('parameters', [('Dictionary', "{'max_depth': [1, 2, 3, 4]}")])]),
                ],
                [
                    (0, 'data', 2, 'data'),
                    (0, 'target', 2, 'target'),
                    (1, 'model', 2, 'model')
                ]
            ))
        self.assertEqual(len(scores[0]), 4)
        self.assertTrue(parameters[0]['max_depth'], 2)

    def test_pipeline(self):
        with intercept_results(Iris, 'target', Predict, 'prediction') as (y_true, y_pred):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('preprocessing|StandardScaler', identifier, []),
                    ('feature_selection|SelectKBest', identifier,
                        [('k', [('Integer', '2')])]),
                    ('classifiers|LinearSVC', identifier, []),
                    ('Pipeline', identifier, []),
                    ('Predict', identifier, [])
                ],
                [
                    # feed data to pipeline
                    (0, 'data', 4, 'training_data'),
                    (0, 'target', 4, 'training_target'),
                    # put models in pipeline
                    (1, 'model', 4, 'model1'),
                    (2, 'model', 4, 'model2'),
                    (3, 'model', 4, 'model3'),
                    # predict using pipeline
                    (4, 'model', 5, 'model'),
                    (0, 'data', 5, 'data')
                ]
            ))
            y_true, y_pred = np.array(y_true[0]), np.array(y_pred[0])
            self.assertEqual(y_true.shape, y_pred.shape)
            self.assertTrue(np.mean(y_true == y_pred) > .8)

    def test_nested_cross_validation(self):
        with intercept_results(CrossValScore, 'scores') as (scores, ):
            self.assertFalse(execute(
                [
                    ('datasets|Iris', identifier, []),
                    ('classifiers|DecisionTreeClassifier', identifier, []),
                    ('GridSearchCV', identifier,
                     [('parameters', [('Dictionary', "{'max_depth': [1, 2, 3, 4]}")])]),
                    ('cross-validation|CrossValScore', identifier, [])
                ],
                [
                    (0, 'data', 3, 'data'),
                    (0, 'target', 3, 'target'),
                    (1, 'model', 2, 'model'),
                    (2, 'model', 3, 'model')
                ]
            ))
        self.assertEqual(len(scores[0]), 3)
        self.assertTrue(np.mean(scores[0]) > .8)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.domain import DBLoopExec, DBLoopIteration


class LoopExec(DBLoopExec):
    """Class that stores info for logging a loop.
    """

    def __copy__(self):
        return self.do_copy()

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBLoopExec.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = LoopExec
        return cp

    @staticmethod
    def convert(_loop_exec):
        if _loop_exec.__class__ == LoopExec:
            return
        _loop_exec.__class__ = LoopExec
        for iteration in _loop_exec.loop_iterations:
            LoopIteration.convert(iteration)

    ##########################################################################
    # Properties

    id = DBLoopExec.db_id
    ts_start = DBLoopIteration.db_ts_start
    ts_end = DBLoopIteration.db_ts_end

    def _get_loop_iterations(self):
        return self.db_loop_iterations
    def _set_loop_iterations(self, iterations):
        self.db_loop_iterations = iterations
    loop_iterations = property(_get_loop_iterations, _set_loop_iterations)
    def add_loop_iteration(self, iteration):
        self.db_add_loop_iteration(iteration)


class LoopIteration(DBLoopIteration):
    """Class that stores info for logging a single iteration of a loop.
    """

    def __copy__(self):
        return self.do_copy()

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBLoopIteration.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = LoopIteration
        return cp

    @staticmethod
    def convert(_loop_iteration):
        from vistrails.core.log.module_exec import ModuleExec
        from vistrails.core.log.group_exec import GroupExec

        if _loop_iteration.__class__ == LoopIteration:
            return
        _loop_iteration.__class__ = LoopIteration
        for item_exec in _loop_iteration.item_execs:
            if item_exec.vtType == ModuleExec.vtType:
                ModuleExec.convert(item_exec)
            elif item_exec.vtType == GroupExec.vtType:
                GroupExec.convert(item_exec)
            elif item_exec.vtType == LoopExec.vtType:
                LoopExec.convert(item_exec)

    ##########################################################################
    # Properties

    id = DBLoopIteration.db_id
    ts_start = DBLoopIteration.db_ts_start
    ts_end = DBLoopIteration.db_ts_end
    completed = DBLoopIteration.db_completed
    error = DBLoopIteration.db_error
    iteration = DBLoopIteration.db_iteration

    def _get_duration(self):
        if self.db_ts_end is not None:
            return self.db_ts_end - self.db_ts_start
        return None
    duration = property(_get_duration)

    def _get_item_execs(self):
        return self.db_item_execs
    def _set_item_execs(self, item_execs):
        self.db_item_execs = item_execs
    item_execs = property(_get_item_execs, _set_item_execs)
    def add_item_exec(self, item_exec):
        self.db_add_item_exec(item_exec)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.gui.common_widgets import QToolWindowInterface

class QVistrailsPaletteInterface(QToolWindowInterface):
    def __init__(self):
        QToolWindowInterface.__init__(self)
        self.controller = None
        self.title = None
        self.p_id = None
        self.main_window = None
 
    @classmethod
    def instance(klass):
        if not hasattr(klass, '_instance'):
            klass._instance = klass()
        return klass._instance

    def toolWindow(self):
        tool_window = QToolWindowInterface.toolWindow(self)
        return tool_window

    def set_id(self, p_id):
        self.p_id = p_id

    def get_id(self):
        return self.p_id

    def set_title(self, title):
        self.setWindowTitle(title)

    def get_title(self):
        return str(self.windowTitle())

    def set_action(self, action):
        self.action = action
        self.connect(self.toolWindow(), 
                     QtCore.SIGNAL("visibilityChanged(bool)"),
                     self.visibility_changed)

    def get_action(self):
        return self.action
        # return self.toolWindow().toggleViewAction()

    def get_action_tuple(self):
        return (self.get_title(), self.get_title(), 
                {'checkable': True, 
                 'checked': self.toolWindow().isVisible(),
                 'callback': self.set_visible})

    def set_main_window(self, mw):
        self.main_window = mw

    def set_visible(self, enabled):
        #print "set_visible ", self, enabled
        if hasattr(self, 'main_window') and self.main_window is not None:
            self.main_window.show()
            self.main_window.raise_()

        if enabled:
            self.toolWindow().show()
            self.toolWindow().raise_()
            
    def set_pin_status(self, pin_status):
        self.toolWindow().setPinStatus(pin_status)
    
    def get_pin_status(self):
        return self.toolWindow().pinStatus
    
    def set_controller(self, controller):
        self.controller = controller

    def get_controller(self):
        return self.controller

    def set_module(self, module):
        self.module = module
        
    def get_module(self):
        return self.module

    def set_descriptor(self, descriptor):
        self.descriptor = descriptor

    def get_descriptor(self):
        return self.descriptor

    def visibility_changed(self, visible):
        self.action.setChecked(visible)



#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

import os.path
import subprocess
import sys
import urllib2

this_dir = os.path.dirname(os.path.abspath(__file__))

DOWNLOAD_URL = "http://www.vistrails.org/usersguide/dev/html/VisTrails.pdf"
SAVE_TO = os.path.abspath(sys.argv[1]) if len(sys.argv) > 1 else this_dir

# http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
def which(program):
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def download_usersguide():
    print "Downloading usersguide from", DOWNLOAD_URL
    response = urllib2.urlopen(DOWNLOAD_URL)
    filename = os.path.join(SAVE_TO,
                            DOWNLOAD_URL.split('/')[-1])
    f = open(filename, 'wb')
    f.write(response.read())
    f.close()
    
if __name__ == "__main__":
    if which('sphinx-build') is None:
        print "Sphinx is not installed!"
        download_usersguide()
    elif which('pdflatex') is None:
        print "pdflatex is not installed!"
        download_usersguide()
    else:
        cwd = os.getcwd()
        os.chdir(this_dir)
        # Build usersguide
        proc = subprocess.Popen(['./build_usersguide.py', SAVE_TO],
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        proc.wait()
        if proc.returncode != 0:
            print "ERROR: building usersguide failed."
            if proc.stdout:
                print proc.stdout.readlines()        
            sys.exit(1)
        os.chdir(this_dir)




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" Module used when running  vistrails uninteractively """
from __future__ import absolute_import, division
import os.path
import unittest

from vistrails.core.application import is_running_gui
from vistrails.core.configuration import get_vistrails_configuration
import vistrails.core.interpreter.default
import vistrails.core.db.io
from vistrails.core.db.io import load_vistrail
from vistrails.core.db.locator import XMLFileLocator, ZIPFileLocator
from vistrails.core import debug
import vistrails.core.interpreter.cached
from vistrails.core.vistrail.job import Workflow as JobWorkflow
import vistrails.core.vistrail.pipeline
from vistrails.core.utils import VistrailsInternalError
from vistrails.core.vistrail.controller import VistrailController
import vistrails.core.packagemanager
import vistrails.core.system
import vistrails.core.vistrail
import vistrails.db


################################################################################
    
def run_and_get_results(w_list, parameters='', output_dir=None, 
                        update_vistrail=True, extra_info=None, 
                        reason='Console Mode Execution'):
    """run_and_get_results(w_list: list of (locator, version), parameters: str,
                           output_dir:str, update_vistrail: boolean,
                           extra_info:dict)
    Run all workflows in w_list, and returns an interpreter result object.
    version can be a tag name or a version id.
    
    """
    elements = parameters.split("$&$")
    aliases = {}
    params = []
    result = []
    for locator, workflow in w_list:
        (v, abstractions , thumbnails, mashups)  = load_vistrail(locator)
        controller = VistrailController(v, locator, abstractions, thumbnails,
                                        mashups, auto_save=update_vistrail)
        if isinstance(workflow, basestring):
            version = v.get_version_number(workflow)
        elif isinstance(workflow, (int, long)):
            version = workflow
        elif workflow is None:
            version = controller.get_latest_version_in_graph()
        else:
            msg = "Invalid version tag or number: %s" % workflow
            raise VistrailsInternalError(msg)
        controller.change_selected_version(version)
        
        for e in elements:
            pos = e.find("=")
            if pos != -1:
                key = e[:pos].strip()
                value = e[pos+1:].strip()
            
                if controller.current_pipeline.has_alias(key):
                    aliases[key] = value
                elif 'mashup_id' in extra_info:
                    # new-style mashups can have aliases not existing in pipeline
                    for mashuptrail in mashups:
                        if mashuptrail.vtVersion == version:
                            mashup = mashuptrail.getMashup(extra_info['mashup_id'])
                            c = mashup.getAliasByName(key).component
                            params.append((c.vttype, c.vtid, value))

        if output_dir is not None and controller.current_pipeline is not None:
            # FIXME DAK: why is this always done?!? there is a flag for it...
            if is_running_gui():
                controller.updatePipelineScene()
                base_fname = "%s_%s_pipeline.pdf" % (locator.short_filename, version)
                filename = os.path.join(output_dir, base_fname)
                controller.current_pipeline_scene.saveToPDF(filename)
            else:
                debug.critical("Cannot save pipeline figure when not "
                               "running in gui mode")
            base_fname = "%s_%s_pipeline.xml" % (locator.short_filename, version)
            filename = os.path.join(output_dir, base_fname)
            vistrails.core.db.io.save_workflow(controller.current_pipeline, filename)
        if not update_vistrail:
            conf = get_vistrails_configuration()
            if conf.has('thumbs'):
                conf.thumbs.autoSave = False
        
        jobMonitor = controller.jobMonitor
        current_workflow = jobMonitor.currentWorkflow()
        if not current_workflow:
            for job in jobMonitor.workflows.itervalues():
                try:
                    job_version = int(job.version)
                except ValueError:
                    job_version =  v.get_version_number(job.version)
                if version == job_version:
                    current_workflow = job
                    jobMonitor.startWorkflow(job)
            if not current_workflow:
                current_workflow = JobWorkflow(version)
                jobMonitor.startWorkflow(current_workflow)

        try:
            (results, _) = \
            controller.execute_current_workflow(custom_aliases=aliases,
                                                custom_params=params,
                                                extra_info=extra_info,
                                                reason=reason)
        finally:
            jobMonitor.finishWorkflow()
        new_version = controller.current_version
        if new_version != version:
            debug.log("Version '%s' (%s) was upgraded. The actual "
                      "version executed was %s" % (
                      workflow, version, new_version))
        run = results[0]
        run.workflow_info = (locator.name, new_version)
        run.pipeline = controller.current_pipeline

        if update_vistrail:
            controller.write_vistrail(locator)
        result.append(run)
        if current_workflow.jobs:
            if current_workflow.completed():
                run.job = "COMPLETED"
            else:
                run.job = "RUNNING: %s" % current_workflow.id
                for job in current_workflow.jobs.itervalues():
                    if not job.finished:
                        run.job += "\n  %s %s %s" % (job.start, job.name, job.description())
            print run.job
    return result

################################################################################

def get_wf_graph(w_list, output_dir=None, pdf=False):
    """run_and_get_results(w_list: list of (locator, version), 
                           output_dir:str, pdf:bool)
    Load all workflows in wf_list and dump their graph to output_dir.
    
    """
    result = []
    if is_running_gui():
        from vistrails.gui.vistrail_controller import VistrailController as \
             GUIVistrailController
        for locator, workflow in w_list:
            try:
                (v, abstractions , thumbnails, mashups)  = load_vistrail(locator)
                controller = GUIVistrailController(v, locator, abstractions, 
                                                   thumbnails, mashups,
                                                   auto_save=False)
                version = None
                if isinstance(workflow, basestring):
                    version = v.get_version_number(workflow)
                elif isinstance(workflow, (int, long)):
                    version = workflow
                elif workflow is None:
                    version = controller.get_latest_version_in_graph()
                else:
                    msg = "Invalid version tag or number: %s" % workflow
                    raise VistrailsInternalError(msg)
            
                if (output_dir is not None and 
                    controller.current_pipeline is not None):
                    controller.updatePipelineScene()
                    if pdf:
                        base_fname = "%s_%s_pipeline.pdf" % \
                                     (locator.short_filename, version)
                        filename = os.path.join(output_dir, base_fname)
                        controller.current_pipeline_scene.saveToPDF(filename)
                    else:
                        base_fname = "%s_%s_pipeline.png" % \
                                     (locator.short_filename, version)
                        filename = os.path.join(output_dir, base_fname)
                        controller.current_pipeline_scene.saveToPNG(filename)
                    result.append((True, ""))
            except Exception, e:
                result.append((False, debug.format_exception(e)))
    else:
        error_str = "Cannot save pipeline figure when not " \
            "running in gui mode"
        debug.critical(error_str)
        result.append((False, error_str))
    return result

################################################################################

def get_vt_graph(vt_list, tree_info, pdf=False):
    """get_vt_graph(vt_list: list of locator, tree_info:str)
    Load all vistrails in vt_list and dump their tree to tree_info.
    
    """
    result = []
    if is_running_gui():
        from vistrails.gui.vistrail_controller import VistrailController as \
             GUIVistrailController
        for locator in vt_list:
            try:
                (v, abstractions , thumbnails, mashups)  = load_vistrail(locator)
                controller = GUIVistrailController(v, locator, abstractions, 
                                                   thumbnails, mashups)
                if tree_info is not None:
                    from vistrails.gui.version_view import QVersionTreeView
                    version_view = QVersionTreeView()
                    version_view.scene().setupScene(controller)
                    if pdf:
                        base_fname = "graph_%s.pdf" % locator.short_filename
                        filename = os.path.join(tree_info, base_fname)
                        version_view.scene().saveToPDF(filename)
                    else:
                        base_fname = "graph_%s.png" % locator.short_filename
                        filename = os.path.join(tree_info, base_fname)
                        version_view.scene().saveToPNG(filename)
                    del version_view
                    result.append((True, ""))
            except Exception, e:
                result.append((False, debug.format_exception(e)))
    else:
        error_str = "Cannot save version tree figure when not " \
            "running in gui mode"
        debug.critical(error_str)
        result.append((False, error_str))
    return result

################################################################################

def run(w_list, parameters='', output_dir=None, update_vistrail=True,
        extra_info=None, reason="Console Mode Execution"):
    """run(w_list: list of (locator, version), parameters: str) -> boolean
    Run all workflows in w_list, version can be a tag name or a version id.
    Returns list of errors (empty list if there are no errors)
    """
    all_errors = []
    results = run_and_get_results(w_list, parameters, output_dir, 
                                  update_vistrail,extra_info, reason)
    for result in results:
        (objs, errors, executed) = (result.objects,
                                    result.errors, result.executed)
        for err in sorted(errors.iteritems()):
            all_errors.append(result.workflow_info + err)
    return all_errors

def run_parameter_exploration(locator, pe_id, extra_info = {},
                              reason="Console Mode Parameter Exploration Execution"):
    """run_parameter_exploration(w_list: (locator, version),
                                 pe_id: str/int,
                                 reason: str) -> (pe_id, [error msg])
    Run parameter exploration in w, and returns an interpreter result object.
    version can be a tag name or a version id.
    
    """
    if is_running_gui():
        from vistrails.gui.vistrail_controller import VistrailController as \
             GUIVistrailController
        try:
            (v, abstractions , thumbnails, mashups)  = load_vistrail(locator)
            controller = GUIVistrailController(v, locator, abstractions, 
                                               thumbnails, mashups)
            try:
                pe_id = int(pe_id)
                pe = controller.vistrail.get_paramexp(pe_id)
            except ValueError:
                pe = controller.vistrail.get_named_paramexp(pe_id)
            controller.change_selected_version(pe.action_id)
            controller.executeParameterExploration(pe, extra_info=extra_info,
                                                   showProgress=False)
        except Exception, e:
            return (locator, pe_id,
                    debug.format_exception(e), debug.format_exc())

def run_parameter_explorations(w_list, extra_info = {},
                       reason="Console Mode Parameter Exploration Execution"):
    """run(w_list: list of (locator, pe_id), reason: str) -> boolean
    For each workflow in w_list, run parameter exploration pe_id
    version can be a tag name or a version id.
    Returns list of errors (empty list if there are no errors)
    """
    all_errors = []
    for locator, pe_id in w_list:
        result = run_parameter_exploration(locator, pe_id, reason=reason,
                                           extra_info=extra_info)
        if result:
            all_errors.append(result)
    return all_errors

def cleanup():
    vistrails.core.interpreter.cached.CachedInterpreter.cleanup()

################################################################################
#Testing


class TestConsoleMode(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        manager = vistrails.core.packagemanager.get_package_manager()
        if manager.has_package('org.vistrails.vistrails.console_mode_test'):
            return

        d = {'console_mode_test': 'vistrails.tests.resources.'}
        manager.late_enable_package('console_mode_test',d)

    @classmethod
    def tearDownClass(cls):
        manager = vistrails.core.packagemanager.get_package_manager()
        if manager.has_package('org.vistrails.vistrails.console_mode_test'):
            manager.late_disable_package('console_mode_test')
            
    def test1(self):
        from vistrails.core.modules.basic_modules import StandardOutput
        values = []
        def mycompute(s):
            v = s.get_input('value')
            values.append(v)
        orig_compute = StandardOutput.compute
        StandardOutput.compute = mycompute
        try:
            locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() +
                                     '/tests/resources/dummy.xml')
            result = run([(locator, "int chain")], update_vistrail=False)
            self.assertEqual(len(result), 0)
            self.assertEqual(values, [2])
        finally:
            StandardOutput.compute = orig_compute

    def test_tuple(self):
        from vistrails.core.vistrail.module_param import ModuleParam
        from vistrails.core.vistrail.module_function import ModuleFunction
        from vistrails.core.utils import DummyView
        from vistrails.core.vistrail.module import Module
        import vistrails.db.domain
       
        id_scope = vistrails.db.domain.IdScope()
        interpreter = vistrails.core.interpreter.default.get_default_interpreter()
        v = DummyView()
        p = vistrails.core.vistrail.pipeline.Pipeline()
        params = [ModuleParam(id=id_scope.getNewId(ModuleParam.vtType),
                              pos=0,
                              type='Float',
                              val='2.0',
                              ),
                  ModuleParam(id=id_scope.getNewId(ModuleParam.vtType),
                              pos=1,
                              type='Float',
                              val='2.0',
                              )]
        function = ModuleFunction(id=id_scope.getNewId(ModuleFunction.vtType),
                                  name='input')
        function.add_parameters(params)
        module = Module(id=id_scope.getNewId(Module.vtType),
                           name='TestTupleExecution',
                           package='org.vistrails.vistrails.console_mode_test',
                           version='0.9.1')
        module.add_function(function)

        p.add_module(module)

        interpreter.execute(p,
                            locator=XMLFileLocator('foo'),
                            current_version=1L,
                            view=v)

    def test_python_source(self):
        locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() +
                                 '/tests/resources/pythonsource.xml')
        result = run([(locator,"testPortsAndFail")], update_vistrail=False)
        self.assertEqual(len(result), 0)

    def test_python_source_2(self):
        locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() +
                                 '/tests/resources/pythonsource.xml')
        result = run_and_get_results([(locator, "test_simple_success")],
                                     update_vistrail=False)[0]
        self.assertEquals(len(result.executed), 1)

    def test_dynamic_module_error(self):
        locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() + 
                                 '/tests/resources/dynamic_module_error.xml')
        result = run([(locator, "test")], update_vistrail=False)
        self.assertNotEqual(len(result), 0)

    def test_change_parameter(self):
        locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() + 
                                 '/tests/resources/test_change_vistrail.xml')
        result = run([(locator, "v1")], update_vistrail=False)
        self.assertEqual(len(result), 0)

        result = run([(locator, "v2")], update_vistrail=False)
        self.assertEquals(len(result), 0)

    def test_ticket_73(self):
        # Tests serializing a custom-named module to disk
        locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() + 
                                 '/tests/resources/test_ticket_73.xml')
        v = locator.load()

        import tempfile
        (fd, filename) = tempfile.mkstemp()
        os.close(fd)
        locator = XMLFileLocator(filename)
        try:
            locator.save(v)
        finally:
            os.remove(filename)

if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from datetime import datetime

from vistrails.core.system import strftime, time_strptime
from vistrails.db.domain import DBMashupAction
from vistrails.core.mashup.mashup import Mashup

class Action(DBMashupAction):
    
    def __init__(self, id, prevId, user='', mashup=None, date=None):
        DBMashupAction.__init__(self, id, prevId, date, user, mashup)
        
    id = DBMashupAction.db_id
    prevId = DBMashupAction.db_prevId
    parent_id = DBMashupAction.db_prevId
    user = DBMashupAction.db_user
    mashup = DBMashupAction.db_mashup
        
    def _get_date(self):
        if self.db_date is not None:
            return strftime(self.db_date, '%d %b %Y %H:%M:%S')
        return strftime(datetime(1900,1,1), '%d %b %Y %H:%M:%S')

    def _set_date(self, date):
        if isinstance(date, datetime):
            self.db_date = date
        elif isinstance(date, basestring) and date.strip() != '':
            newDate = datetime(*time_strptime(date, '%d %b %Y %H:%M:%S')[0:6])
            self.db_date = newDate
    date = property(_get_date, _set_date)
        
    @staticmethod
    def convert(_action):
        if _action.__class__ == Action:
            return
        _action.__class__ = Action
        Mashup.convert(_action.mashup)
         
    def __copy__(self):
        return Action.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBMashupAction.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = Action
        Mashup.convert(cp.mashup)
        return cp
    
#    def toXml(self, node=None):
#        """toXml(node: ElementTree.Element) -> ElementTree.Element
#           writes itself to xml
#        """
#
#        if node is None:
#            node = ElementTree.Element('action')
#        
#        #set attributes
#        node.set('id', self.convert_to_str(self.id,'long'))
#        node.set('parent_id', self.convert_to_str(self.parent_id,'long'))
#        node.set('user', self.convert_to_str(self.user,'str'))
#        node.set('date', self.convert_to_str(self._date,'datetime'))
#        mnode = ElementTree.SubElement(node, 'mashup')
#        self.mashup.toXml(mnode)
#        return node
#    
#    @staticmethod
#    def fromXml(node):
#        if node.tag != 'action':
#            debug.debug("node.tag != 'action'")
#            return None
#        #read attributes
#        data = node.get('id', None)
#        id = Action.convert_from_str(data, 'long')
#        data = node.get('parent_id', None)
#        parent_id = Action.convert_from_str(data, 'long')
#        data = node.get('user', None)
#        user = Action.convert_from_str(data, 'str')
#        data = node.get('date', None)
#        date = Action.convert_from_str(data, 'datetime')
#        child = node.getchildren()[0]
#        if child.tag == 'mashup':
#            mashup = Mashup.fromXml(child)
#        return Action(id=id, parent_id=parent_id, mashup=mashup, user=user,
#                      date=date)
        
    ##########################################################################
    # Operators
    
    def __str__(self):
        """ __str__() -> str - Returns a string representation of itself """
        
        msg = "<<type='%s' id='%s' parent_id='%s' date='%s' user='%s'>>"
        return msg % (type(self),
                      self.id,
                      self.prevId,
                      self.db_date,
                      self.user)
        
    def __eq__(self, other):
        """ __eq__(other: Alias) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.mashup != other.mashup:
            return False
        return True

    def __ne__(self, other):
        """ __ne__(other: Component) -> boolean
        Returns True if self and other don't have the same attributes. 
        Used by !=  operator. 
        
        """
        return not self.__eq__(other)
################################################################################
            
import unittest
from vistrails.db.domain import IdScope
import copy

class TestAction(unittest.TestCase):
    def create_action(self, id_scope=IdScope()):
        from vistrails.core.mashup.component import Component
        from vistrails.core.mashup.alias import Alias
        from vistrails.core.mashup.mashup import Mashup
        c1 = Component(id=id_scope.getNewId('mashup_component'),
                          vttype='parameter', param_id=15L, 
                          parent_vttype='function', parent_id=3L, mid=4L,
                          type='String', value='test', p_pos=0, pos=1, 
                          strvaluelist='test1,test2', widget="text")
        a1 = Alias(id=id_scope.getNewId('mashup_alias'), name='alias1', component=c1)
        
        m = Mashup(id=id_scope.getNewId('mashup'), name='mashup1', vtid='empty.vt', 
                   version=15L, alias_list=[a1])
        action = Action(id=id_scope.getNewId('mashup_action'),
                        prevId=0L,
                        date=datetime(2007,11,18),
                        mashup=m)
        return action
    
    def test_copy(self):
        id_scope = IdScope()
        a1 = self.create_action(id_scope)
        a2 = copy.copy(a1)
        self.assertEquals(a1, a2)
        self.assertEquals(a1.id, a2.id)
        a3 = a1.do_copy(True, id_scope, {})
        self.assertEquals(a1, a3)
        self.assertNotEquals(a1.id, a3.id)

#    def test_serialization(self):
#        a1 = self.create_action()
#        node = a1.toXml()
#        a2 = Action.fromXml(node)
#        self.assertEquals(a1, a2)
#        self.assertEquals(a1.id, a2.id)
        
    def test_str(self):
        a1 = self.create_action()
        str(a1)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This common widgets using on the interface of VisTrails. These are
only simple widgets in term of coding and additional features. It
should have no interaction with VisTrail core"""
from __future__ import division

import os

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import pyqtSlot, pyqtSignal
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.modules.constant_configuration import StandardConstantWidget
from vistrails.core.system import systemType, set_vistrails_data_directory
################################################################################

class QToolWindow(QtGui.QDockWidget):
    """
    QToolWindow is a floating-dockable widget. It also keeps track of
    its widget window title to update the tool window accordingly
    
    """
    def __init__(self, widget=None, parent=None):
        """ QToolWindow(parent: QWidget) -> QToolWindow
        Construct a floating, dockable widget
        
        """
        QtGui.QDockWidget.__init__(self, parent)
        self.setFeatures(QtGui.QDockWidget.AllDockWidgetFeatures)  
        self.mwindow = QtGui.QMainWindow(self)
        self.centralwidget = widget
        self.mwindow.setWindowFlags(QtCore.Qt.Widget) 
        self.mwindow.setCentralWidget(widget)     
        self.setWidget(self.mwindow)
        self.createToolBar()
        if widget:
            self.setWindowTitle(widget.windowTitle())
        self.pinStatus = False
        self.monitorWindowTitle(widget)
        
        self.connect(self, QtCore.SIGNAL("topLevelChanged(bool)"),
                     self.setDefaultPinStatus)
             
    def createToolBar(self):
        self.toolbar = QtGui.QToolBar(self.mwindow)
        self.pinButton = QtGui.QAction(CurrentTheme.UNPINNED_PALETTE_ICON,
                                       "", self.toolbar,checkable=True,
                                       checked=False,
                                       toggled=self.pinStatusChanged)
        
        self.pinButton.setToolTip("Pin this on the Tab Bar")
        spacer = QtGui.QWidget()
        spacer.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding, 
                             QtGui.QSizePolicy.Preferred)
        self.toolbar.addWidget(spacer)
        self.toolbar.addAction(self.pinButton)
        self.pinAction = self.pinButton
        self.toolbar.setFloatable(False)
        self.toolbar.setMovable(False)
        self.toolbar.setIconSize(QtCore.QSize(16,16))
        self.mwindow.addToolBar(self.toolbar)
                   
    def setDefaultPinStatus(self, topLevel):
        # Fixes QTBUG-30276
        if self.acceptDrops():
            self.setAcceptDrops(False)
            self.setAcceptDrops(True)
        if topLevel:
            self.setPinStatus(False)
            self.pinButton.setEnabled(False)
        else:
            self.pinButton.setEnabled(True)
        
    def pinStatusChanged(self, pinStatus):
        self.pinStatus = pinStatus
        self.updateButtonIcon(pinStatus)
        
    def updateButtonIcon(self, on):
        if on:
            self.pinButton.setIcon(CurrentTheme.PINNED_PALETTE_ICON)
            self.pinButton.setToolTip("Unpin this from the the Tab Bar")
        else:
            self.pinButton.setIcon(CurrentTheme.UNPINNED_PALETTE_ICON)
            self.pinButton.setToolTip("Pin this on the Tab Bar")
            
    def setPinStatus(self, pinStatus):
        self.pinStatus = pinStatus
        self.pinButton.setChecked(pinStatus)
        self.updateButtonIcon(pinStatus)
        
    def monitorWindowTitle(self, widget):
        """ monitorWindowTitle(widget: QWidget) -> None        
        Watching window title changed on widget and use it as a window
        title on this tool window
        
        """
        if widget:
            widget.installEventFilter(self)

    def eventFilter(self, object, event):
        """ eventFilter(object: QObject, event: QEvent) -> bool
        Filter window title change event to change the tool window title
        
        """
        if event.type()==QtCore.QEvent.WindowTitleChange:
            self.setWindowTitle(object.windowTitle())
        elif event.type()==QtCore.QEvent.Close:
            object.removeEventFilter(self)
        return QtGui.QDockWidget.eventFilter(self, object, event)
        # return super(QToolWindow, self).eventFilter(object, event)                    
            
        
class QToolWindowInterface(object):
    """
    QToolWindowInterface can be co-inherited in any class to allow the
    inherited class to switch to be contained in a window
    
    """
    def toolWindow(self):
        """ toolWindow() -> None        
        Return the tool window and set its parent to self.parent()
        while having self as its contained widget
        
        """
        if not hasattr(self, '_toolWindow'):
            self._toolWindow = QToolWindow(self, self.parent and self.parent())
        elif self._toolWindow.centralwidget!=self:
            self._toolWindow.window.setCentralWidget(self)
        return self._toolWindow

    def changeEvent(self, event):
        """ changeEvent(event: QEvent) -> None        
        Make sure to update the tool parent when to match the widget's
        real parent
        
        """
        if (event.type()==QtCore.QEvent.ParentChange and
            hasattr(self, '_toolWindow')):
            if self.parent()!=self._toolWindow:
                self._toolWindow.setParent(self.parent())

    def setToolWindowAcceptDrops(self, value):
        self.toolWindow().setAcceptDrops(value)

###############################################################################

class QDockContainer(QtGui.QMainWindow):
    """
    QDockContainer is a window that can contain dock widgets while
    still be contained in a tool window. It is just a straight
    inheritance from QMainWindow
    
    """
    def __init__(self, parent=None):
        """ QMainWindow(parent: QWidget) -> QMainWindow
        Setup window to have its widget dockable everywhere
        
        """
        QtGui.QMainWindow.__init__(self, parent)
        self.setDockNestingEnabled(True)

###############################################################################

class QSearchTreeWidget(QtGui.QTreeWidget):
    """
    QSearchTreeWidget is just a QTreeWidget with a support function to
    refine itself when searching for some text

    """
    def __init__(self, parent=None):
        """ QSearchTreeWidget(parent: QWidget) -> QSearchTreeWidget
        Set up size policy and header

        """
        QtGui.QTreeWidget.__init__(self, parent)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Expanding)
        self.setRootIsDecorated(True)
        self.setDragEnabled(True)
        self.flags = QtCore.Qt.ItemIsDragEnabled

        self._search_was_empty = True

    def searchItemName(self, name):
        """ searchItemName(name: QString) -> None
        Search and refine the module tree widget to contain only items
        whose name is 'name'
        
        """
        matchedItems = []
        
        def recursiveSetVisible(item, testFunction):
            """ recursiveSetVisible
            Pass through all items of a item
            
            """
            enabled = testFunction(item)

            visible = enabled
            child = item.child
            for childIndex in xrange(item.childCount()):
                visible |= recursiveSetVisible(child(childIndex),
                                               testFunction)

            # if item is hidden or has changed visibility
            if not visible or (item.isHidden() != (not visible)):
                item.setHidden(not visible)
            
            if visible:
                f = item.flags()
                b = f & self.flags
                if enabled:
                    if not b:
                        item.setFlags(f | self.flags)
                elif b:
                    item.setFlags(f & ~self.flags)
                
            return visible

        if str(name)=='':
            testFunction = lambda x: (not hasattr(x, 'is_hidden') or
                                      not x.is_hidden)
            if not self._search_was_empty:
                self.collapseAll()
                self._search_was_empty = True
        else:
            matchedItems = set(self.findItems(name,
                                              QtCore.Qt.MatchContains |
                                              QtCore.Qt.MatchWrap |
                                              QtCore.Qt.MatchRecursive))
            testFunction = matchedItems.__contains__
            if self._search_was_empty:
                self.expandAll()
                self._search_was_empty = False
        for itemIndex in xrange(self.topLevelItemCount()):
            recursiveSetVisible(self.topLevelItem(itemIndex),
                                testFunction)
    
    def mimeData(self, itemList):
        """ mimeData(itemList) -> None        
        Setup the mime data to contain itemList because Qt 4.2.2
        implementation doesn't instantiate QTreeWidgetMimeData
        anywhere as it's supposed to. It must have been a bug...
        
        """
        data = QtGui.QTreeWidget.mimeData(self, itemList)
        data.items = itemList
        return data

    def setMatchedFlags(self, flags):
        """ setMatchedFlags(flags: QItemFlags) -> None Set the flags
        for matched item in the search tree. Parents of matched node
        will be visible with these flags off.
        
        """
        self.flags = flags
    
class QSearchTreeWindow(QtGui.QWidget):
    """
    QSearchTreeWindow contains a search box on top of a tree widget
    for easy search and refine. The subclass has to implement
    createTreeWidget() method to return a tree widget that is also 
    needs to expose searchItemName method

    """
    def __init__(self, parent=None):
        """ QSearchTreeWindow(parent: QWidget) -> QSearchTreeWindow
        Intialize all GUI components
        
        """
        QtGui.QWidget.__init__(self, parent)
        self.setWindowTitle('Search Tree')
        
        vLayout = QtGui.QVBoxLayout(self)
        vLayout.setMargin(0)
        vLayout.setSpacing(0)
        self.setLayout(vLayout)
        
        self.searchBox = QSearchBox(False, True, self)
        vLayout.addWidget(self.searchBox)

        self.treeWidget = self.createTreeWidget()
        vLayout.addWidget(self.treeWidget)
        
        self.connect(self.searchBox,
                     QtCore.SIGNAL('executeIncrementalSearch(QString)'),
                     self.treeWidget.searchItemName)
        self.connect(self.searchBox,
                     QtCore.SIGNAL('executeSearch(QString)'),
                     self.treeWidget.searchItemName)
        self.connect(self.searchBox,
                     QtCore.SIGNAL('resetSearch()'),
                     self.clearTreeWidget)
                     
    def clearTreeWidget(self):
        """ clearTreeWidget():
        Return the default search tree

        """
        self.treeWidget.searchItemName('')

    def createTreeWidget(self):
        """ createTreeWidget() -> QSearchTreeWidget
        Return a default searchable tree widget
        
        """
        return QSearchTreeWidget(self)

class QPromptWidget(QtGui.QLabel):
    """
    QPromptWidget is a widget that will display a prompt text when it
    doesn't have any child visible, or else, it will disappear. This
    is good for drag and drop prompt. The inheritance should call
    setPromptText and showPrompt in appropriate time to show/hide the
    prompt text
    """
    def __init__(self, parent=None):
        """ QPromptWidget(parent: QWidget) -> QPromptWidget
        Set up the font and alignment for the prompt
        
        """
        QtGui.QLabel.__init__(self, parent)        
        self.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.setWordWrap(True)
        self.regularFont = self.font()
        self.promptFont = QtGui.QFont(self.font())
        self.promptFont.setItalic(True)
        self.promptText = ''
        self.promptVisible = False

    def setPromptText(self, text):
        """ setPromptText(text: str) -> None
        Set the prompt text string
        
        """
        self.promptText = text

    def showPrompt(self, show=True):
        """ showPrompt(show: boolean) -> None
        Show/Hide the prompt
        
        """
        if show!=self.promptVisible:
            self.promptVisible = show
            self.repaint(self.rect())
            
    def showPromptByChildren(self):
        """ showPromptByChildren()
        Show/Hide the prompt based on the current state of children
        
        """
        if self.promptText=='':
            self.showPrompt(False)
        else:
            self.showPrompt(self.layout()==None or
                            self.layout().count()==0)            

    def paintEvent(self, event):
        """ paintEvent(event: QPaintEvent) -> None
        Paint the prompt in the center if neccesary
        
        """
        if self.promptVisible:
            painter = QtGui.QPainter(self)
            painter.setFont(self.promptFont)
            painter.drawText(self.rect(),
                             QtCore.Qt.AlignCenter | QtCore.Qt.TextWordWrap,
                             self.promptText)
            painter.end()
        QtGui.QLabel.paintEvent(self, event)
        # super(QPromptWidget, self).paintEvent(event)

class QStringEdit(QtGui.QFrame):
    """
    QStringEdit is a line edit that has an extra button to allow user
    to use a file as the value
    
    """
    def __init__(self, parent=None):
        """ QStringEdit(parent: QWidget) -> QStringEdit
        Create a hbox layout to contain a line edit and a button

        """
        QtGui.QFrame.__init__(self, parent)        
        hLayout = QtGui.QHBoxLayout(self)
        hLayout.setMargin(0)
        hLayout.setSpacing(0)        
        self.setLayout(hLayout)

        self.lineEdit = QtGui.QLineEdit()
        self.lineEdit.setFrame(False)
        self.lineEdit.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                    QtGui.QSizePolicy.Expanding)
        hLayout.addWidget(self.lineEdit)
        self.setFocusProxy(self.lineEdit)

        self.fileButton = QtGui.QToolButton()
        self.fileButton.setText('...')
        self.fileButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                                      QtGui.QSizePolicy.Expanding)
        self.fileButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.fileButton.setAutoFillBackground(True)
        self.connect(self.fileButton, QtCore.SIGNAL('clicked()'),
                     self.insertFileNameDialog)
        hLayout.addWidget(self.fileButton)

    def setText(self, text):
        """ setText(text: QString) -> None
        Overloaded function for setting the line edit text
        
        """
        self.lineEdit.setText(text)

    def text(self):
        """ text() -> QString
        Overloaded function for getting the line edit text
        
        """
        return self.lineEdit.text()

    def selectAll(self):
        """ selectAll() -> None
        Overloaded function for selecting all the text
        
        """
        self.lineEdit.selectAll()

    def setFrame(self, frame):
        """ setFrame(frame: bool) -> None
        Show/Hide the frame of this widget
        
        """
        if frame:
            self.setFrameStyle(QtGui.QFrame.StyledPanel |
                               QtGui.QFrame.Plain)
        else:
            self.setFrameStyle(QtGui.QFrame.NoFrame)

    def insertFileNameDialog(self):
        """ insertFileNameDialog() -> None
        Allow user to insert a file name as a value to the string
        
        """
        fileName = QtGui.QFileDialog.getOpenFileName(self,
                                                     'Use Filename '
                                                     'as Value...',
                                                     self.text(),
                                                     'All files '
                                                     '(*.*)')
        if fileName:
            self.setText(fileName)

###############################################################################

class QSearchEditBox(QtGui.QComboBox):
    def __init__(self, incremental=True, parent=None):
        QtGui.QComboBox.__init__(self, parent)
        self.setEditable(True)
        self.setInsertPolicy(QtGui.QComboBox.InsertAtTop)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Fixed)
        regexp = QtCore.QRegExp("\S.*")
        self.setDuplicatesEnabled(False)
        validator = QtGui.QRegExpValidator(regexp, self)
        self.setValidator(validator)
        self.addItem('Clear Recent Searches')
        item = self.model().item(0, 0)
        font = QtGui.QFont(item.font())
        font.setItalic(True)
        item.setFont(font)
        self.is_incremental = incremental

    def keyPressEvent(self, e):
        if e.key() in (QtCore.Qt.Key_Return,QtCore.Qt.Key_Enter):
            if self.currentText():
                if not self.is_incremental:
                    self.emit(QtCore.SIGNAL('executeSearch(QString)'),  
                              self.currentText())
                self.insertItem(0, self.currentText())
            else:
                self.emit(QtCore.SIGNAL('resetText()'))
            return
        QtGui.QComboBox.keyPressEvent(self, e)
        
###############################################################################

class QSearchBox(QtGui.QWidget):
    """ 
    QSearchBox contains a search combo box with a clear button and
    a search icon.

    """
    def __init__(self, refine=True, incremental=True, parent=None):
        """ QSearchBox(parent: QWidget) -> QSearchBox
        Intialize all GUI components
        
        """
        QtGui.QWidget.__init__(self, parent)
        self.setWindowTitle('Search')
        
        hLayout = QtGui.QHBoxLayout(self)
        hLayout.setMargin(0)
        hLayout.setSpacing(2)
        self.setLayout(hLayout)

        self.searchEdit = QSearchEditBox(incremental, self)
        #TODO: Add separator!
        self.searchEdit.clearEditText()

        if refine:
            self.actionGroup = QtGui.QActionGroup(self)
            self.searchAction = QtGui.QAction('Search', self)
            self.searchAction.setCheckable(True)
            self.actionGroup.addAction(self.searchAction)
            self.refineAction = QtGui.QAction('Refine', self)
            self.refineAction.setCheckable(True)
            self.actionGroup.addAction(self.refineAction)
            self.searchAction.setChecked(True)

            self.searchMenu = QtGui.QMenu()
            self.searchMenu.addAction(self.searchAction)
            self.searchMenu.addAction(self.refineAction)

            self.searchButton = QtGui.QToolButton(self)
            self.searchButton.setIcon(CurrentTheme.QUERY_ARROW_ICON)
            self.searchButton.setAutoRaise(True)
            self.searchButton.setPopupMode(QtGui.QToolButton.InstantPopup)
            self.searchButton.setMenu(self.searchMenu)
            hLayout.addWidget(self.searchButton)
            self.connect(self.searchAction, QtCore.SIGNAL('triggered()'),
                         self.searchMode)
            self.connect(self.refineAction, QtCore.SIGNAL('triggered()'),
                         self.refineMode)
        else:
            self.searchLabel = QtGui.QLabel(self)
            pix = CurrentTheme.QUERY_VIEW_ICON.pixmap(QtCore.QSize(16,16))
            self.searchLabel.setPixmap(pix)
            self.searchLabel.setAlignment(QtCore.Qt.AlignCenter)
            self.searchLabel.setMargin(4)
            hLayout.addWidget(self.searchLabel)
        
        hLayout.addWidget(self.searchEdit)

        self.resetButton = QtGui.QToolButton(self)
        self.resetButton.setIcon(QtGui.QIcon(
                self.style().standardPixmap(QtGui.QStyle.SP_DialogCloseButton)))
        self.resetButton.setIconSize(QtCore.QSize(12,12))
        self.resetButton.setAutoRaise(True)
        self.resetButton.setEnabled(False)
        hLayout.addWidget(self.resetButton)
        self.manualResetEnabled = False

        self.connect(self.resetButton, QtCore.SIGNAL('clicked()'),
                     self.resetSearch)
        self.connect(self.searchEdit, QtCore.SIGNAL('activated(int)'),
                     self.executeSearch)
        self.connect(self.searchEdit, QtCore.SIGNAL('resetText'),
                     self.resetSearch)
        self.connect(self.searchEdit, QtCore.SIGNAL('executeSearch(QString)'),
                     self.executeTextSearch)
        if incremental:
            self.connect(self.searchEdit, 
                         QtCore.SIGNAL('editTextChanged(QString)'),
                         self.executeIncrementalSearch)
        else:
            self.connect(self.searchEdit,
                         QtCore.SIGNAL('editTextChanged(QString)'),
                         self.resetToggle)

    def resetSearch(self):
        """
        resetSearch() -> None
        Emit a signal to clear the search.

        """
        self.searchEdit.clearEditText()
        self.resetButton.setEnabled(False)
        self.manualResetEnabled = False
        self.emit(QtCore.SIGNAL('resetSearch()'))

    def clearSearch(self):
        """ clearSearch() -> None
        Clear the edit text without emitting resetSearch() signal
        This is for when the search is rest from the version view and
        the signal are already taken care of

        """
        self.searchEdit.clearEditText()
        self.resetButton.setEnabled(False)
        self.manualResetEnabled = False

    def searchMode(self):
        """
        searchMode() -> None

        """
        self.emit(QtCore.SIGNAL('refineMode(bool)'), False) 
    
    def refineMode(self):
        """
        refineMode() -> None

        """
        self.emit(QtCore.SIGNAL('refineMode(bool)'), True) 

    def resetToggle(self, text):
        self.resetButton.setEnabled((str(text) != '') or 
                                    self.manualResetEnabled)

    def executeIncrementalSearch(self, text):
        """
        executeIncrementalSearch(text: QString) -> None
        The text is changing, so update the search.

        """
        self.resetButton.setEnabled((str(text)!='') or
                                    self.manualResetEnabled)
        self.emit(QtCore.SIGNAL('executeIncrementalSearch(QString)'), text)

    def executeTextSearch(self, text):
        self.emit(QtCore.SIGNAL('executeSearch(QString)'),
                  text)

    def executeSearch(self, index):
        """
        executeSearch(index: int) -> None
        The text is finished changing or a different item was selected.

        """
        count = self.searchEdit.count() 
        if index == count-1: 
            for i in xrange(count-1): 
                self.searchEdit.removeItem(0) 
            self.resetSearch() 
        else: 
            self.resetButton.setEnabled(True) 
            self.emit(QtCore.SIGNAL('executeSearch(QString)'),  
                      self.searchEdit.currentText())

    def getCurrentText(self):
        return str(self.searchEdit.currentText())

    def setManualResetEnabled(self, boolVal):
        self.manualResetEnabled = boolVal
        self.resetButton.setEnabled((self.getCurrentText() != '') or
                                    self.manualResetEnabled)

###############################################################################

class QMouseTabBar(QtGui.QTabBar):
    """QMouseTabBar is a QTabBar that emits a signal when a tab
    receives a mouse event. For now only doubleclick events are
    emitted."""
    #signals
    tabDoubleClicked = pyqtSignal(int,QtCore.QPoint)
    
    def __init__(self, parent=None):
        QtGui.QTabBar.__init__(self, parent)
        
    def mouseDoubleClickEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            tab_idx = self.tabAt(event.pos())
            if tab_idx != -1:
                self.tabDoubleClicked.emit(tab_idx, event.pos())
        QtGui.QTabBar.mouseDoubleClickEvent(self, event)

###############################################################################

class QDockPushButton(QtGui.QPushButton):
    """QDockPushButton is a button to be used inside QDockWidgets. It will 
    set the minimum height on Mac so it looks nice on both Mac and Windows"""
    def __init__(self, text, parent=None):
        QtGui.QPushButton.__init__(self, text, parent) 
        if systemType in ['Darwin']:
            self.setMinimumHeight(32)

class QPathChooserToolButton(QtGui.QToolButton):
    """
    QPathChooserToolButton is a toolbar button that opens a browser for
    paths.  The lineEdit is updated with the pathname that is selected.
    
    emits pathChanged when the path is changed

    """
    pathChanged = pyqtSignal()

    def __init__(self, parent=None, lineEdit=None, toolTip=None,
                 defaultPath=None):
        """
        PathChooserToolButton(parent: QWidget, 
                              lineEdit: StandardConstantWidget) ->
                 PathChooserToolButton

        """
        QtGui.QToolButton.__init__(self, parent)
        self.setIcon(QtGui.QIcon(
                self.style().standardPixmap(QtGui.QStyle.SP_DirOpenIcon)))
        self.setIconSize(QtCore.QSize(12,12))
        if toolTip is None:
            toolTip = 'Open a path chooser'
        self.defaultPath = defaultPath
        self.setToolTip(toolTip)
        self.setAutoRaise(True)
        self.lineEdit = lineEdit
        self.connect(self,
                     QtCore.SIGNAL('clicked()'),
                     self.runDialog)

    def setPath(self, path):
        """
        setPath() -> None

        """
        if self.lineEdit and path:
            self.lineEdit.setText(path)
            self.pathChanged.emit()
    
    def getDefaultText(self):
        return self.lineEdit.text() or self.defaultPath

    def openChooser(self):
        path = QtGui.QFileDialog.getOpenFileName(self,
                                                 'Select Path...',
                                                 self.getDefaultText(),
                                                 'All files '
                                                 '(*.*)')
        return self.setDataDirectory(path)

    def runDialog(self):
        path = self.openChooser()
        self.setPath(path)

    def setDataDirectory(self, path):
        if path:
            absPath = os.path.abspath(str(QtCore.QFile.encodeName(path)))
            dirName = os.path.dirname(absPath)
            set_vistrails_data_directory(dirName)
            return absPath
        return path

class QFileChooserToolButton(QPathChooserToolButton):
    def __init__(self, parent=None, lineEdit=None, toolTip=None,
                 defaultPath=None):
        if toolTip is None:
            toolTip = "Open a file chooser dialog"
        QPathChooserToolButton.__init__(self, parent, lineEdit, toolTip,
                                        defaultPath)

    def openChooser(self):
        path = QtGui.QFileDialog.getOpenFileName(self,
                                                 'Select File...',
                                                 self.getDefaultText(),
                                                 'All files '
                                                 '(*.*)')
        return self.setDataDirectory(path)

class QDirectoryChooserToolButton(QPathChooserToolButton):
    def __init__(self, parent=None, lineEdit=None, toolTip=None,
                 defaultPath=None):
        if toolTip is None:
            toolTip = "Open a directory chooser dialog"
        QPathChooserToolButton.__init__(self, parent, lineEdit, toolTip,
                                       defaultPath)

    def openChooser(self):
        path = QtGui.QFileDialog.getExistingDirectory(self,
                                                      'Select Directory...',
                                                      self.getDefaultText())
        return self.setDataDirectory(path)

class QOutputPathChooserToolButton(QPathChooserToolButton):
    def __init__(self, parent=None, lineEdit=None, toolTip=None,
                 defaultPath=None):
        if toolTip is None:
            toolTip = "Open a path chooser dialog"
        QPathChooserToolButton.__init__(self, parent, lineEdit, toolTip,
                                       defaultPath)
    
    def openChooser(self):
        path = QtGui.QFileDialog.getSaveFileName(self,
                                                 'Select Output Location...',
                                                 self.getDefaultText(),
                                                 'All files (*.*)')
        return self.setDataDirectory(path)
    
    



from __future__ import division

from PyQt4 import QtCore, QtGui

from vistrails.core.system import get_vistrails_basic_pkg_id
from vistrails.gui.modules.module_configure import \
    StandardModuleConfigurationWidget

from .identifiers import identifier
from vistrails.core.modules.basic_modules import List


class Entry(QtGui.QWidget):
    remove = QtCore.pyqtSignal()
    changed = QtCore.pyqtSignal()

    def __init__(self, name):
        QtGui.QWidget.__init__(self)
        layout = QtGui.QHBoxLayout()
        self.setLayout(layout)
        self.lineedit = QtGui.QLineEdit(name)
        layout.addWidget(QtGui.QLabel(self.prefix))
        layout.addWidget(self.lineedit)
        layout.addStretch()
        remove_button = QtGui.QPushButton("Remove port")
        remove_button.setSizePolicy(QtGui.QSizePolicy.Fixed,
                                    QtGui.QSizePolicy.Fixed)
        layout.addWidget(remove_button)

        self.connect(remove_button, QtCore.SIGNAL('clicked()'),
                     self.remove)
        self.connect(self.lineedit, QtCore.SIGNAL('textEdited(const QString &)'),
                     self.changed)

    @property
    def name(self):
        return self.lineedit.text()


class TableEntry(Entry):
    prefix = "Table input"


class ColumnEntry(Entry):
    prefix = "Single column entry"


class BuildTableWidget(StandardModuleConfigurationWidget):
    """
    Configuration widget allowing to create the ports of the BuildTable module.
    """
    def __init__(self, module, controller, parent=None):
        StandardModuleConfigurationWidget.__init__(self, module,
                                                   controller, parent)

        # Window title
        self.setWindowTitle("Build table configuration")

        central_layout = QtGui.QVBoxLayout()
        central_layout.setMargin(0)
        central_layout.setSpacing(0)
        self.setLayout(central_layout)

        self._scroll_area = QtGui.QScrollArea()
        inner_widget = QtGui.QWidget()
        self._list_layout = QtGui.QVBoxLayout()
        scroll_layout = QtGui.QVBoxLayout()
        scroll_layout.addLayout(self._list_layout)
        scroll_layout.addStretch()
        inner_widget.setLayout(scroll_layout)
        self._scroll_area.setVerticalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOn)
        self._scroll_area.setWidget(inner_widget)
        self._scroll_area.setWidgetResizable(True)
        central_layout.addWidget(self._scroll_area)

        add_buttons = QtGui.QHBoxLayout()
        central_layout.addLayout(add_buttons)
        add_table = QtGui.QPushButton("Add a whole table")
        self.connect(add_table, QtCore.SIGNAL('clicked()'),
                     self.add_table)
        add_buttons.addWidget(add_table)
        add_column = QtGui.QPushButton("Add a list as a single column")
        self.connect(add_column, QtCore.SIGNAL('clicked()'),
                     self.add_column)
        add_buttons.addWidget(add_column)

        self.createButtons()

        self.createEntries()

    def add_item(self, item):
        self._list_layout.addWidget(item)
        self.connect(item, QtCore.SIGNAL('remove()'),
                     lambda: item.deleteLater())
        self.connect(item, QtCore.SIGNAL('changed()'),
                     self.updateState)

    def add_table(self):
        self.add_item(TableEntry(
                "Table #%d" % (self._list_layout.count() + 1)))
        self.updateState()

    def add_column(self):
        self.add_item(ColumnEntry(
                "Column #%d" % (self._list_layout.count() + 1)))
        self.updateState()

    def createButtons(self):
        """ createButtons() -> None
        Create and connect signals to Ok & Cancel button

        """
        buttonLayout = QtGui.QHBoxLayout()
        buttonLayout.setMargin(5)
        self.saveButton = QtGui.QPushButton("&Save", self)
        self.saveButton.setFixedWidth(100)
        self.saveButton.setEnabled(False)
        buttonLayout.addWidget(self.saveButton)
        self.resetButton = QtGui.QPushButton("&Reset", self)
        self.resetButton.setFixedWidth(100)
        self.resetButton.setEnabled(False)
        buttonLayout.addWidget(self.resetButton)
        self.layout().addLayout(buttonLayout)
        self.connect(self.saveButton, QtCore.SIGNAL('clicked(bool)'),
                     self.saveTriggered)
        self.connect(self.resetButton, QtCore.SIGNAL('clicked(bool)'),
                     self.resetTriggered)

    def saveTriggered(self, checked = False):
        """ saveTriggered(checked: bool) -> None
        Update vistrail controller and module when the user click Ok

        """
        if self.updateVistrail():
            self.saveButton.setEnabled(False)
            self.resetButton.setEnabled(False)
            self.state_changed = False
            self.emit(QtCore.SIGNAL('stateChanged'))
            self.emit(QtCore.SIGNAL('doneConfigure'), self.module.id)

    def closeEvent(self, event):
        self.askToSaveChanges()
        event.accept()

    def getCurrentPorts(self):
        current_ports = []
        for port_spec in self.module.input_port_specs:
            is_table = port_spec.signature[0][0] is not List
            current_ports.append((port_spec.name, is_table))
        return current_ports

    def updateVistrail(self):
        """ updateVistrail() -> None
        Update Vistrail to contain changes in the port table

        """
        table_sig = '(%s:Table)' % identifier
        list_sig = '(%s:List)' % get_vistrails_basic_pkg_id()
        seen_new_ports = set()
        current_ports = dict(self.getCurrentPorts())
        add_ports = []
        delete_ports = []
        for i in xrange(self._list_layout.count()):
            widget = self._list_layout.itemAt(i).widget()
            is_table = isinstance(widget, TableEntry)
            name = widget.name

            if name in seen_new_ports:
                QtGui.QMessageBox.critical(
                        self,
                        "Duplicated port name",
                        "There is several input ports with name %r" % name)
                return
            seen_new_ports.add(name)

            if name in current_ports:
                old_is_table = current_ports.pop(name)
                if is_table == old_is_table:
                    continue
                delete_ports.append(('input', name))

            sigstring = table_sig if is_table else list_sig
            add_ports.append(('input', name,
                              sigstring, -1))

        delete_ports.extend(('input', unseen_port)
                            for unseen_port in current_ports.iterkeys())

        self.controller.update_ports(self.module.id, delete_ports, add_ports)

        return True

    def createEntries(self):
        for name, is_table in self.getCurrentPorts():
            if is_table:
                self.add_item(TableEntry(name))
            else:
                self.add_item(ColumnEntry(name))

    def resetTriggered(self, checked = False):
        for i in xrange(self._list_layout.count()):
            self._list_layout.itemAt(i).widget().deleteLater()

        self.createEntries()

        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)
        self.state_changed = False
        self.emit(QtCore.SIGNAL('stateChanged'))

    def updateState(self):
        self.saveButton.setEnabled(True)
        self.resetButton.setEnabled(True)
        if not self.state_changed:
            self.state_changed = True
            self.emit(QtCore.SIGNAL('stateChanged'))



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

################################################################################

from __future__ import division

class Queue(object):

    def __init__(self, capacity=8):
        """ Queue(capacity: int) -> Queue
        Initialize the queue with a default capacity of zero in size

        """
        self.__buffer = [None] * capacity
        self.__capacity = capacity
        self.__size = 0
        self.__begin = 0
        self.__end = 0
        
    def __len__(self):
        """ __len__() -> int
        Compute the number of elements when using the built-in len()
        and return an int

        """
        l = self.__end - self.__begin
        if l < 0:
            return self.__capacity + l
        else:
            return l
        
    def front(self):
        """ front() -> element type
        Get the front of the queue and return an element type

        """
        return self.__buffer[self.__begin]
    
    def back(self):
        """ back() -> element type
        Get the back (last) element of the queue and return an element type

        """
        return self.__buffer[(self.__end + self.__capacity - 1) %
                             self.__capacity]
    
    def push(self, obj):
        """ push(obj: element type) -> None
        Push obj onto the back queue and return nothing

        """
        # olen = len(self)
        if (self.__end + 1) % self.__capacity == self.__begin:
            self.__rebuffer(self.__capacity * 2)
        self.__buffer[self.__end] = obj
        self.__end += 1
        if self.__end == self.__capacity:
            self.__end = 0
        # nlen = len(self)
        # assert nlen == olen + 1

    class EmptyQueue(Exception):
        pass
    
    def pop(self):
        """ pop() -> element type
        Pop the front element of the queue and return an element type

        """
        # olen = len(self)
        if len(self) == 0:
            raise self.EmptyQueue()
        r = self.__buffer[self.__begin]
        self.__buffer[self.__begin] = None
        self.__begin += 1
        if self.__begin == self.__capacity:
            self.__begin = 0
        if self.__capacity > 8 and self.__capacity >= 4 * len(self):
            self.__rebuffer(self.__capacity // 2)
        # nlen = len(self)
        # assert olen == nlen + 1
        return r

    def capacity(self):
        return self.__capacity

    def __rebuffer(self, newcapacity):
        """ __rebuffer(newcapacity: int) -> none
        Reallocate the internal buffer to fit at newcapacity elements and
        return nothing

        """
        nb = [None] * newcapacity
        if self.__begin < self.__end:
            l = self.__end - self.__begin
            nb[0:l] = self.__buffer[self.__begin:self.__end]
            self.__begin = 0
            self.__end = l
        else:
            l1 = self.__capacity - self.__begin
            l2 = l1 + self.__end
            nb[0:l1] = self.__buffer[self.__begin:]
            nb[l1:l2] = self.__buffer[:self.__end]
            self.__begin = 0
            self.__end = l2
        self.__buffer = nb
        self.__capacity = newcapacity
        
    def __str__(self):
        """ __str__() -> str
        Format the queue for serialization and return a string

        """
        return str(self.__buffer)

################################################################################

import unittest
import random

class TestQueue(unittest.TestCase):
    
    def test_basic(self):
        """Test push/pop operations"""
        q = Queue()
        q.push(1)
        q.push(2)
        q.push(3)
        q.push(4)
        self.assertEquals(q.pop(), 1)
        self.assertEquals(q.pop(), 2)
        self.assertEquals(q.pop(), 3)
        self.assertEquals(q.pop(), 4)
        
    def test_expand_basic(self):
        """Test if the queue is expanding its capacity right with push()"""
        q = Queue()
        for i in xrange(512):
            assert len(q) == i
            q.push(i)
            assert len(q) == i+1
        self.assertEquals(len(q), 512)
        for i in xrange(512):
            assert len(q) == 512 - i
            self.assertEquals(q.pop(), i)
            assert len(q) == 511 - i
            
    def test_expand_contract(self):
        """Test if the queue is expanding and contracting with push()/pop()"""
        def run_it():
            pushed = 0
            popped = 0
            q = Queue()
            for t in xrange(100):
                for i in xrange(100):
                    # Test expand with high probability
                    a = random.choice([0,0,0,0,0,1])
                    if (a == 0) or (len(q) == 0):
                        olen = len(q)
                        q.push(pushed)
                        nlen = len(q)
                        assert nlen == olen+1, "push invariant"
                        pushed += 1
                    else:
                        olen = len(q)
                        v = q.pop()
                        nlen = len(q)
                        assert olen == nlen+1, "pop invariant"
                        self.assertEquals(v, popped)
                        popped += 1
                for i in xrange(100):
                    # Test contract with high probability
                    a = random.choice([1,1,1,1,1,0])
                    if (a == 0) or (len(q) == 0):
                        olen = len(q)
                        q.push(pushed)
                        nlen = len(q)
                        assert nlen == olen+1, "push invariant"
                        pushed += 1
                    else:
                        olen = len(q)
                        v = q.pop()
                        nlen = len(q)
                        assert olen == nlen+1, "pop invariant"
                        self.assertEquals(v, popped)
                        popped += 1
        # run test 100 times
        for i in xrange(1):
            run_it()

    def test_pop(self):
        x = Queue()
        try:
            x.pop()
        except x.EmptyQueue:
            pass
        else:
            assert False

if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.db.locator import BaseLocator
from vistrails.core.system import strftime
from datetime import datetime

class Entity(object):
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    def __init__(self):
        self.parent = None
        self.children = []
        self.image_fnames = []
        self.was_updated = False
        self.is_open = False

    def load(self, *args):
        (self.id, 
         _, 
         self.name,
         self.user,
         self.mod_time,
         self.create_time,
         self.size,
         self.description,
         self.url) = args
        self.mod_time = self.timeval(self.mod_time)
        self.create_time = self.timeval(self.create_time)


    def save(self):
        return (self.id,
                self.type_id,
                self.name,
                self.user,
                strftime(self.mod_time, self.DATE_FORMAT),
                strftime(self.create_time, self.DATE_FORMAT),
                self.size,
                self.description,
                self.url)

    def _get_start_date(self):
        return self.create_time
    start_date = property(_get_start_date)

    def _get_end_date(self):
        return self.mod_time
    end_date = property(_get_end_date)

    def now(self):
        return datetime.now()

    def timeval(self, time):
        try:
            return datetime.strptime(time, self.DATE_FORMAT)
        except ValueError:
            # try old format
            try:
                return datetime.strptime(time, '%d %b %Y %H:%M:%S')
            except ValueError:
                # locale or other error
                return self.now()

#     # returns string
#     def get_name(self):
#         return self.name

#     # returns datetime
#     def get_mod_time(self):
#         raise RuntimeError("Method is abstract")

#     # returns datetime
#     def get_create_time(self):
#         raise RuntimeError("Method is abstract")
    
#     # returns string
#     # FIXME: perhaps this should be a User object at some point
#     def get_user(self):
#         raise RuntimeError("Method is abstract")
    
#     # returns tuple (<entity_type>, <entity_id>)
#     def get_id(self):
#         raise RuntimeError("Method is abstract")

#     # returns integer
#     def get_size(self):
#         raise RuntimeError("Method is abstract")

#     def get_description(self):
#         raise RuntimeError("Method is abstract")
    
#     # returns possibly empty list of Entity objects
#     def get_children(self):
#         raise RuntimeError("Method is abstract")

#     # returns list of strings representing paths
#     # FIXME: should this be uris for database access?
#     def get_image_fnames(self):
#         raise RuntimeError("Method is abstract")
    
    # returns boolean, True if search input is satisfied else False
    def match(self, search):
        raise RuntimeError("Method is abstract")

    def locator(self):
        locator = BaseLocator.from_url(self.url)
        return locator




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from datetime import datetime

from vistrails.core.system import strftime, time_strptime
from vistrails.db.domain import DBMashupActionAnnotation

class ActionAnnotation(DBMashupActionAnnotation):
    def __init__(self, id, action_id, key=None, value=None, user=None, date=None):
        DBMashupActionAnnotation.__init__(self, id, key, value, action_id, date, 
                                          user)
    id = DBMashupActionAnnotation.db_id
    action_id = DBMashupActionAnnotation.db_action_id
    key = DBMashupActionAnnotation.db_key
    value = DBMashupActionAnnotation.db_value
    user = DBMashupActionAnnotation.db_user
        
    def _get_date(self):
        if self.db_date is not None:
            return strftime(self.db_date, '%d %b %Y %H:%M:%S')
        return strftime(datetime(1900,1,1), '%d %b %Y %H:%M:%S')

    def _set_date(self, date):
        if isinstance(date, datetime):
            self.db_date = date
        elif isinstance(date, basestring) and date.strip() != '':
            newDate = datetime(*time_strptime(date, '%d %b %Y %H:%M:%S')[0:6])
            self.db_date = newDate
    date = property(_get_date, _set_date)
    
    @staticmethod
    def convert(_annotation):
        _annotation.__class__ = ActionAnnotation
        
    def __copy__(self):
        return ActionAnnotation.do_copy(self)
    
    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBMashupActionAnnotation.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ActionAnnotation
        return cp
    
#    def toXml(self, node=None):
#        """toXml(node: ElementTree.Element) -> ElementTree.Element
#           writes itself to xml
#        """
#
#        if node is None:
#            node = ElementTree.Element('actionAnnotation')
#        
#        #set attributes
#        node.set('id', self.convert_to_str(self.id,'long'))
#        node.set('action_id', self.convert_to_str(self.action_id,'long'))
#        node.set('user', self.convert_to_str(self.user,'str'))
#        node.set('key', self.convert_to_str(self.key,'str'))
#        node.set('value', self.convert_to_str(self.value,'str'))
#        node.set('date', self.convert_to_str(self._date,'datetime'))
#        return node
#    
#    @staticmethod
#    def fromXml(node):
#        if node.tag != 'actionAnnotation':
#            debug.debug("node.tag != 'actionAnnotation'")
#            return None
#        #read attributes
#        data = node.get('id', None)
#        id = Action.convert_from_str(data, 'long')
#        data = node.get('action_id', None)
#        action_id = Action.convert_from_str(data, 'long')
#        data = node.get('key', None)
#        key = Action.convert_from_str(data, 'str')
#        data = node.get('value', None)
#        value = Action.convert_from_str(data, 'str')
#        data = node.get('user', None)
#        user = Action.convert_from_str(data, 'str')
#        data = node.get('date', None)
#        date = Action.convert_from_str(data, 'datetime')
#        return ActionAnnotation(id=id, action_id=action_id, key=key, value=value,
#                      user=user, date=date)
    
    ##########################################################################
    # Operators
    
    def __str__(self):
        """__str__() -> str - Returns a string representation of an
        mashup_actionAnnotation object.

        """
        rep = ("<mashup_actionAnnotation id=%s action_id=%s key=%s value=%s "
               "date=%s user=%s</annotation>")
        return  rep % (str(self.id), str(self.action_id), str(self.key), 
                       str(self.value), str(self.date), str(self.user))

    def __eq__(self, other):
        """ __eq__(other: mashup_actionAnnotation) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.key != other.key:
            return False
        if self.value != other.value:
            return False
        if self.action_id != other.action_id:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)
        
################################################################################
import unittest
from vistrails.db.domain import IdScope
import copy

class TestActionAnnotation(unittest.TestCase):
    def create_annotation(self, id_scope=IdScope()):

        annotation = \
            ActionAnnotation(id=id_scope.getNewId('mashup_actionAnnotation'),
                             key='akey', action_id=1L,
                             value='some value', user='test')
        return annotation

    def test_copy(self):
        id_scope = IdScope()
        a1 = self.create_annotation(id_scope)
        a2 = copy.copy(a1)
        self.assertEquals(a1, a2)
        self.assertEquals(a1.id, a2.id)
        a3 = a1.do_copy(True, id_scope, {})
        self.assertEquals(a1, a3)
        self.assertNotEquals(a1.id, a3.id)

#    def test_serialization(self):
#        a1 = self.create_annotation()
#        node = a1.toXml()
#        a2 = ActionAnnotation.fromXml(node)
#        self.assertEquals(a1, a2)
#        self.assertEquals(a1.id, a2.id)

    def test_str(self):
        a1 = self.create_annotation()
        str(a1)



#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""To use this script, run VisTrails with the old version of a package enabled, and choose the Export -> Registry to XML... and save the file.  Then, enable the new version of the package and again export the registry to a second XML file.  Then, run this script with the package identifier, the old registry's filename, and the new registry's filename as arguments."""

import os
import sys
# put the vistrails code on the python path
sys.path.append(os.path.join(os.path.dirname(os.path.dirname(
                os.path.abspath(__file__))), "vistrails"))

import vistrails.core.application
from vistrails.core.modules.module_registry import ModuleRegistry
from vistrails.db.services.io import open_registry_from_xml

def get_module_name(d):
    if d.namespace:
        return "%s|%s" % (d.namespace, d.name)
    return d.name

def diff_package(pkg_id, reg_fname1, reg_fname2):
    reg1 = open_registry_from_xml(reg_fname1)
    ModuleRegistry.convert(reg1)
    reg2 = open_registry_from_xml(reg_fname2)
    ModuleRegistry.convert(reg2)
    
    pkg1 = reg1.get_package_by_name(pkg_id)
    pkg2 = reg2.get_package_by_name(pkg_id)

    d2_modules_dict = dict(((d2.identifier, d2.name, d2.namespace), d2)
                           for d2 in pkg2.descriptor_list)
    for d1 in pkg1.descriptor_list:
        if reg2.has_descriptor_with_name(pkg_id, d1.name, d1.namespace):
            d2 = reg2.get_descriptor_by_name(pkg_id, d1.name, d1.namespace)
            d1_port_specs = {}
            for ps in d1.port_specs_list:
                d1_port_specs[(ps.name, ps.type)] = ps.sigstring
            d2_port_specs = {}
            for ps in d2.port_specs_list:
                d2_port_specs[(ps.name, ps.type)] = ps.sigstring
            d2_port_specs_set = set(d2_port_specs.keys())
            for ps_id, sig1 in d1_port_specs.iteritems():
                if ps_id not in d2_port_specs:
                    print "added %s port: %s:%s" % \
                        (ps_id[1], get_module_name(d1), ps_id[0])
                    continue
                d2_port_specs_set.discard(ps_id)
                if sig1 != d2_port_specs[ps_id]:
                    print "changed %s port: %s:%s" % \
                        (ps_id[1], get_module_name(d1), ps_id[0])
                    print "  %s -> %s" % (sig1, d2_port_specs[ps_id])
                else:
                    # equal
                    pass
            for ps_id in d2_port_specs_set:
                print "deleted %s port: %s:%s" % \
                    (ps_id[1], get_module_name(d1), ps_id[0])
            del d2_modules_dict[(d1.identifier, d1.name, d1.namespace)]
        else:
            print "deleted module: %s" % get_module_name(d1)
    for d2 in d2_modules_dict.itervalues():
        print "added module: %s" % get_module_name(d2)
    
if __name__ == '__main__':
    if len(sys.argv) < 4:
        print "Usage: %s %s <pkg_identifer> <old_registry> <new_registry>" % \
            (sys.executable, sys.argv[0])
        sys.exit(71)
    vistrails.core.application.init()
    diff_package(*sys.argv[1:])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
from vistrails.core import debug

from vistrails.db.versions.v0_9_5.domain import DBVistrail, DBWorkflow, DBLog, \
    DBRegistry, DBModuleExec, DBGroupExec

def translateVistrail(_vistrail):
    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBGroup': {'workflow': update_workflow}}
    vistrail = DBVistrail.update_version(_vistrail, translate_dict)
    vistrail.db_version = '0.9.5'
    return vistrail

def translateWorkflow(_workflow):
    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBGroup': {'workflow': update_workflow}}
    workflow = DBWorkflow.update_version(_workflow, translate_dict)
    workflow.db_version = '0.9.5'
    return workflow

def translateLog(_log):
    def update_items(old_obj, translate_dict):
        new_items = []
        for obj in old_obj.db_item_execs:
            if obj.vtType == 'module_exec':
                new_items.append(DBModuleExec.update_version(obj, translate_dict))
            elif obj.vtType == 'group_exec':
                new_items.append(DBGroupExec.update_version(obj, translate_dict))
            elif obj.vtType == 'loop_exec':
                new_items.append(DBLoopExec.update_version(obj, translate_dict))

        return new_items

    def update_exec_items(which, old_obj, translate_dict, catch_loops=False):
        new_items = []
        for obj in old_obj.db_item_execs:
            if catch_loops and obj.vtType == 'loop_exec':
                debug.warning("Cannot translate loop_exec inside of loop_exec")
            if obj.vtType == which:
                if obj.vtType == 'module_exec':
                    new_items.append(DBModuleExec.update_version(obj, translate_dict))
                elif obj.vtType == 'group_exec':
                    new_items.append(DBGroupExec.update_version(obj, translate_dict))
                elif obj.vtType == 'loop_exec':
                    new_items.append(DBLoopExec.update_version(obj, translate_dict))
        return new_items

    def update_group_module_execs(old_obj, translate_dict):
        return update_exec_items('module_exec', old_obj, translate_dict)

    def update_group_group_execs(old_obj, translate_dict):
        return update_exec_items('group_exec', old_obj, translate_dict)

    def update_group_loop_execs(old_obj, translate_dict):
        return update_exec_items('loop_exec', old_obj, translate_dict)

    def update_loop_module_execs(old_obj, translate_dict):
        return update_exec_tems('module_exec', old_obj, translate_dict, True)

    def update_loop_group_execs(old_obj, translate_dict):
        return update_exec_items('group_exec', old_obj, translate_dict)

    translate_dict = {'DBWorkflowExec': {'items': update_items},
                      'DBGroupExec': {'group_execs': update_group_group_execs,
                                      'module_execs': update_group_module_execs,
                                      'loop_execs': update_group_loop_execs},
                      'DBLoopExec': {'group_execs': update_loop_group_execs,
                                     'module_execs': update_loop_module_execs}}
    log = DBLog.update_version(_log, translate_dict)
    log.db_version = '0.9.5'
    return log

def translateRegistry(_registry):
    translate_dict = {}
    registry = DBRegistry.update_version(_registry, translate_dict)
    registry.db_version = '0.9.5'
    return registry



#!/usr/bin/env python

# This example demonstrates how to use the vtkPointWidget to probe a
# dataset and then color the probed point (represented as a cone) as
# per the probed value and orient it as per the vector.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Start by loading some data.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()

# The plane widget is used probe the dataset.
pointWidget = vtk.vtkPointWidget()
pointWidget.SetInput(pl3d.GetOutput())
pointWidget.AllOff()
pointWidget.PlaceWidget()
point = vtk.vtkPolyData()
pointWidget.GetPolyData(point)

probe = vtk.vtkProbeFilter()
probe.SetInput(point)
probe.SetSource(pl3d.GetOutput())

# create glyph
cone = vtk.vtkConeSource()
cone.SetResolution(16)

inspector = vtk.vtkDataSetInspector()
inspector.SetInputConnection(pl3d.GetOutputPort())

pythoncalc = vtk.PythonCalc()
pythoncalc.value1(inspector.GetLength())
pythoncalc.value2(0.1)
pythoncalc.op('*')

glyph = vtk.vtkGlyph3D()
glyph.SetInputConnection(probe.GetOutputPort())
glyph.SetSource(cone.GetOutput())
glyph.SetVectorModeToUseVector()
glyph.SetScaleModeToDataScalingOff()
# glyph.SetScaleFactor(pl3d.GetOutput().GetLength()*0.1)
glyph.SetScaleFactor(pythoncalc.value())
glyphMapper = vtk.vtkPolyDataMapper()
glyphMapper.SetInputConnection(glyph.GetOutputPort())
glyphActor = vtk.vtkActor()
glyphActor.SetMapper(glyphMapper)
glyphActor.VisibilityOff()

# An outline is shown for context.
outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Actually set the color and orientation of the probe.
def BeginInteraction(obj, event):
    global point, glyphActor
    obj.GetPolyData(point)
    glyphActor.VisibilityOn() 

def ProbeData(obj, event):
    obj.GetPolyData(point) 

# Associate the line widget with the interactor
pointWidget.SetInteractor(iren)
pointWidget.AddObserver("EnableEvent", BeginInteraction)
pointWidget.AddObserver("StartInteractionEvent", BeginInteraction)
pointWidget.AddObserver("InteractionEvent", ProbeData)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddActor(glyphActor)

ren.SetBackground(1, 1, 1)
renWin.SetSize(300, 300)
ren.SetBackground(0.1, 0.2, 0.4)

cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(3.95297, 50)
cam1.SetFocalPoint(9.71821, 0.458166, 29.3999)
cam1.SetPosition(2.7439, -37.3196, 38.7167)
cam1.SetViewUp(-0.16123, 0.264271, 0.950876)

iren.Initialize()
renWin.Render()
iren.Start()



# pragma: no testimport

from __future__ import division

import sys


if __name__ == '__main__':
    args = sys.argv[1:]

    if len(args) != 10:
        sys.stderr.write("Wrong number of parameters\n")
        sys.exit(1)

    fin = args.pop(0)
    fout = args.pop(0)

    if args != ['-t', '42', '-l', 'a', '-l', 'b', '-l', 'c']:
        sys.stderr.write("Wrong parameters\n")
        sys.exit(1)

    stdin = sys.stdin.readline()
    if stdin != 'some line\n':
        sys.stderr.write("Wrong line on stdin\n")
        sys.exit(1)

    if not fin.endswith('.cltest'):
        sys.stderr.write("Wrong extension for f_in '%s'\n" % fin)
        sys.exit(1)

    with open(fin, 'r') as fpin:
        fcin = fpin.read()
    if fcin != 'this is a\ntest':
        sys.stderr.write("Wrong content in f_in '%s'\n" % fin)
        sys.exit(1)

    if not fout.endswith('.cltest'):
        sys.stderr.write("Wrong extension for f_out '%s'\n" % fout)
        sys.exit(1)

    with open(fout, 'wb') as fpout:
        fpout.write('ok\nmessage received')

    sys.stdout.write("program output here")

    sys.exit(0)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from datetime import datetime
import os

from vistrails.core.modules.basic_modules import Directory, File, Path, \
    PathObject
from vistrails.core.modules.config import IPort, OPort, ModuleSettings
from vistrails.core.modules.vistrails_module import Module, ModuleError

from .common import KEY_TYPE, TYPE_OUTPUT, \
    KEY_SIGNATURE, KEY_TIME, KEY_WORKFLOW, KEY_MODULE_ID, get_default_store
from .queries import Metadata


class PersistedPath(Module):
    """Records a file in the file store.
    """

    _input_ports = [
            IPort('path', Path),
            IPort('metadata', Metadata, optional=True)]
    _output_ports = [
            OPort('path', Path)]

    _cached = None

    def update_upstream(self):
        """A modified version of the update_upstream method.

        Only updates upstream if the file is not found in the store.
        """
        if not hasattr(self, 'signature'):
            raise ModuleError(self, "Module has no signature")
        file_store = get_default_store()
        entries = file_store.query({KEY_SIGNATURE: self.signature})
        best = None
        for entry in entries:
            if best is None or entry[KEY_TIME] > best[KEY_TIME]:
                best = entry
        if best is not None:
            self._cached = best.filename
        else:
            super(PersistedPath, self).update_upstream()

    def compute(self):
        if self._cached is not None:
            self._set_result(self._cached)
        else:
            file_store = get_default_store()
            newpath = self.get_input('path').name
            self.check_path_type(newpath)
            metadata = self.get_input_list('metadata')
            metadata = dict(m.metadata for m in metadata)
            metadata[KEY_TYPE] = TYPE_OUTPUT
            metadata[KEY_TIME] = datetime.strftime(datetime.utcnow(),
                                                   '%Y-%m-%d %H:%M:%S')
            metadata[KEY_SIGNATURE] = self.signature
            locator = self.moduleInfo.get('locator')
            if locator is not None:
                metadata[KEY_WORKFLOW] = "%s:%s" % (
                        locator.name,
                        self.moduleInfo['version'])
            metadata[KEY_MODULE_ID] = self.moduleInfo['moduleId']
            entry = file_store.add(newpath, metadata)
            self.annotate({'added_file': entry['hash']})
            self._set_result(entry.filename)

    def check_path_type(self, path):
        pass

    def _set_result(self, path):
        self.set_output('path', PathObject(path))


class PersistedFile(PersistedPath):
    """Records a file in the file store.
    """

    _input_ports = [
            IPort('path', File),
            IPort('metadata', Metadata, optional=True)]
    _output_ports = [
            OPort('path', File)]
    _settings = ModuleSettings(configure_widget=
            'vistrails.packages.persistent_archive.widgets:SetMetadataWidget')

    def check_path_type(self, path):
        if not os.path.isfile(path):
            raise ModuleError(self, "Path is not a file")


class PersistedDir(PersistedPath):
    """Records a directory in the file store.
    """

    _input_ports = [
            IPort('path', Directory),
            IPort('metadata', Metadata, optional=True)]
    _output_ports = [
            OPort('path', Directory)]
    _settings = ModuleSettings(configure_widget=
            'vistrails.packages.persistent_archive.widgets:SetMetadataWidget')

    def check_path_type(self, path):
        if not os.path.isdir(path):
            raise ModuleError(self, "Path is not a directory")
